from rest_framework import serializers as rf


def raw_tag(item):
    '''Returns item tag without namespace'''
    return item.tag.partition('}')[2] or item.tag


def lxml_to_dict(node):
    '''Convert an lxml.etree node tree into a dict recursively'''
    storage = {}

    for child in node.iterchildren():
        key = raw_tag(child)
        storage[key] = child.text if child.text else lxml_to_dict(child)

    return storage


def deque_to_dict(data):
    return {raw_tag(item): lxml_to_dict(item) for item in data}


class DequeToDictMixin:
    def to_internal_value(self, data):
        return deque_to_dict(data)


class ENSSerializer(DequeToDictMixin, rf.Serializer):
    is_error = False

    mrn = rf.CharField(source='HEAHEA.DocNumHEA5')


class EXSSerializer(DequeToDictMixin, rf.Serializer):
    is_error = False

    mrn = rf.CharField(source='HEAHEA.DocNumHEA5')
    item_number_involved = rf.IntegerField(source='RISANA.IteNumInvRKA1')
    customs_intervention_code = rf.CharField(source='RISANA.CusIntCodRKA1')


class UnknownResponseSerializer(rf.Serializer):
    is_error = True

    def __init__(self, *args, **kwargs):
        kwargs.pop('data') if 'data' in kwargs else None
        data = {}
        super().__init__(self, *args, data=data, **kwargs)

    def to_representation(self, obj):
        return {'reason': 'Unknown AEAT response'}


class ENSFunctionalErrorSerializer(DequeToDictMixin, rf.Serializer):
    is_error = True

    code = rf.CharField(source='FUNERRER1.ErrReaER13', default='')
    type = rf.CharField(source='FUNERRER1.ErrTypER11')
    pointer = rf.CharField(source='FUNERRER1.ErrPoiER12')
    reason = rf.CharField(source='FUNERRER1.OriAttValER14', default='No error message supplied')


def parse_xsd(data):
    try:
        body = data.find('.//soapenv:Body', namespaces=data.nsmap)
    except AttributeError:
        body = data

    # Try V2 Style
    try:
        xsd = body[0].nsmap[None]
    except (IndexError, KeyError):
        pass
    else:
        print("Estilo V2", xsd)
        return xsd

    # Try V4 Style
    try:
        xsd = body[0].nsmap['ie']
    except (IndexError, KeyError):
        pass
    else:
        print("Estilo V5", xsd)
        return xsd


def get_class_for_aeat_response(data):
    xsd = parse_xsd(data)

    ens = 'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/aden/enswsv5/'
    exs = 'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/adrx/ws/'

    return {
        f'{ens}IE328V5Sal.xsd': ENSSerializer,
        f'{ens}IE304V5Sal.xsd': ENSSerializer,
        f'{ens}IE316V5Sal.xsd': ENSFunctionalErrorSerializer,
        f'{exs}IE628V2Sal.xsd': EXSSerializer,
    }.get(xsd, UnknownResponseSerializer)



import datetime as dt

import factory


class ENSPresentationHeader(factory.Factory):
    class Meta:
        model = dict

    RefNumHEA4 = 'LRN000000041'
    TraModAtBorHEA76 = '5'
    IdeOfMeaOfTraCroHEA85 = '1111111'
    IdeOfMeaOfTraCroHEA85LNG = 'ES'
    TotNumOfIteHEA305 = '3'
    TotNumOfPacHEA306 = '50'
    TotGroMasHEA307 = '10'
    DecPlaHEA394 = 'Madrid'
    DecPlaHEA394LNG = 'ES'
    SpeCirIndHEA1 = 'A'
    TraChaMetOfPayHEA1 = 'C'
    ComRefNumHEA = 'a828rt'
    ConRefNumHEA = '7777b'
    PlaLoaGOOITE334 = 'ESMadrid'
    PlaLoaGOOITE334LNG = 'ES'
    PlaUnlGOOITE334 = 'ESSegovia'
    CodPlUnHEA357LNG = 'ES'
    DecDatTimHEA114 = factory.LazyAttribute(lambda x: dt.datetime.now())


class ENSModificationHeader(ENSPresentationHeader):
    class Meta:
        model = dict

    DocNumHEA5 = 'mrn_number_xyz'
    AmdPlaHEA598 = 'Barcelona'
    DatTimAmeHEA113 = factory.LazyAttribute(lambda x: dt.datetime.now())


class EXSHeader(factory.Factory):
    class Meta:
        model = dict

    RefNumHEA4 = 'LRN000000041'
    TotNumOfIteHEA305 = '3'
    TotNumOfPacHEA306 = '50'
    TotGroMasHEA307 = '10'
    DecPlaHEA394 = 'Madrid'
    SpeCirIndHEA1 = 'A'
    TraChaMetOfPayHEA1 = 'C'
    ComRefNumHEA = 'a828rt'
    DecDatTimHEA114 = factory.LazyAttribute(lambda x: dt.datetime.now())
    CusSubPlaHEA66 = '4611ZZZ999'


class TraderConsignor(factory.Factory):
    class Meta:
        model = dict

    NamCO17 = 'JUAN CARLOS'
    StrAndNumCO122 = 'Almansa'
    PosCodCO123 = '28007'
    CitCO124 = 'Madrid'
    CouCO125 = 'ES'
    TINCO159 = 'ESA08005688'


class TraderConsignee(factory.Factory):
    class Meta:
        model = dict

    NamCE17 = 'LUÍS'
    StrAndNumCE122 = 'Cruz'
    PosCodCE123 = '28005'
    CitCE124 = 'Madrid'
    CouCE125 = 'ES'
    TINCE159 = 'ESA08005688'


class ProducedDocumentsCertificates(factory.Factory):
    class Meta:
        model = dict

    DocTypDC21 = 'Y022'
    DocRefDC23 = 'ESAEOC1'


class SpecialMentions(factory.Factory):
    class Meta:
        model = dict

    AddInfCodMT23 = 'xxx'


class CommodityCode(factory.Factory):
    class Meta:
        model = dict

    ComNomCMD1 = '123456'


class Container(factory.Factory):
    class Meta:
        model = dict

    ConNumNR21 = '1'


class Package(factory.Factory):
    class Meta:
        model = dict

    KinOfPacGS23 = 'NE'
    NumOfPacGS24 = '0'
    NumOfPieGS25 = '10'
    # MarNumOfPacGSL21 = 'PAQUETES1'
    MarNumOfPacGSL21LNG = 'ES'


class GoodsItem(factory.Factory):
    class Meta:
        model = dict

    IteNumGDS7 = '1'
    GooDesGDS23 = 'DESCRIPCION  PARTIDA1'
    GooDesGDS23LNG = 'ES'
    GroMasGDS46 = '100'
    ComRefNumGIM1 = 'REFERENCIACOMER1'

    SPEMENMT2 = factory.SubFactory(SpecialMentions)
    COMCODGODITM = factory.SubFactory(CommodityCode)

    @factory.post_generation
    def CONNR2(self, create, extracted, **kwargs):
        if extracted:
            # A list of groups were passed in, use them
            self['CONNR2'] = extracted
        else:
            self['CONNR2'] = [Container()]

    @factory.post_generation
    def PRODOCDC2(self, create, extracted, **kwargs):
        if extracted:
            # A list of groups were passed in, use them
            self['PRODOCDC2'] = extracted
        else:
            self['PRODOCDC2'] = [ProducedDocumentsCertificates()]

    @factory.post_generation
    def PACGS2(self, create, extracted, **kwargs):
        if extracted:
            # A list of groups were passed in, use them
            self['PACGS2'] = extracted
        else:
            self['PACGS2'] = [Package()]


class PersonLodgingSummaryDeclaration(factory.Factory):
    class Meta:
        model = dict

    NamPLD1 = 'PEDRO'
    StrAndNumPLD1 = 'CASTELLANA'
    PosCodPLD1 = '28003'
    CitPLD1 = 'MADRID'
    CouCodPLD1 = 'ES'
    PERLODSUMDECLNG = 'ES'
    TINPLD1 = 'ESA08005688'


class SealsIdentity(factory.Factory):
    class Meta:
        model = dict

    SeaIdSEAID530 = '1'
    SeaIdSEAID530LNG = 'ES'


class CustomsOfficeFirstEntry(factory.Factory):
    class Meta:
        model = dict

    RefNumCUSOFFFENT731 = 'ES009999'
    ExpDatOfArrFIRENT733 = factory.LazyAttribute(lambda x:
                                                 dt.datetime.now())


class CustomsOfficeSubsequentEntry(factory.Factory):
    class Meta:
        model = dict

    RefNumSUBENR909 = 'FR000010'


class TraderEntryCarrier(factory.Factory):
    class Meta:
        model = dict

    NamTRACARENT604 = 'ROSA'
    StrNumTRACARENT607 = 'MONCLOA'
    PstCodTRACARENT606 = '28007'
    CtyTRACARENT603 = 'MADRID'
    CouCodTRACARENT605 = 'ES'
    TRACARENT601LNG = 'ES'
    TINTRACARENT602 = 'ESA08005688'


class NotifyParty(factory.Factory):
    class Meta:
        model = dict

    NamNOTPAR672 = 'ROSA'
    StrNumNOTPAR673 = 'MONCLOA'
    PosCodNOTPAR676 = '28007'
    CitNOTPAR674 = 'MADRID'
    CouCodNOTPAR675 = 'ES'
    NOTPAR670LNG = 'ES'
    TINNOTPAR671 = 'ESA08005688'


class BaseMessageMixin(factory.Factory):
    MesSenMES3 = factory.Sequence(lambda n: 'VAT00000%d' % n)
    DatOfPreMES9 = factory.LazyAttribute(lambda x: dt.datetime.now().date())
    TimOfPreMES10 = factory.LazyAttribute(lambda x: dt.datetime.now().time())
    MesIdeMES19 = factory.Sequence(lambda n: 'TESTID000%d' % n)

    HEAHEA = factory.SubFactory(ENSPresentationHeader)
    TRACONCO1 = factory.SubFactory(TraderConsignor)
    TRACONCE1 = factory.SubFactory(TraderConsignee)
    PERLODSUMDEC = factory.SubFactory(PersonLodgingSummaryDeclaration)
    CUSOFFFENT730 = factory.SubFactory(CustomsOfficeFirstEntry)
    TRACARENT601 = factory.SubFactory(TraderEntryCarrier)

    @factory.post_generation
    def GOOITEGDS(self, create, extracted, **kwargs):
        if extracted:
            # A list of groups were passed in, use them
            self['GOOITEGDS'] = extracted
        else:
            self['GOOITEGDS'] = [GoodsItem()]

    @factory.post_generation
    def SEAID529(self, create, extracted, **kwargs):
        if extracted:
            # A list of groups were passed in, use them
            self['SEAID529'] = extracted
        else:
            self['SEAID529'] = [SealsIdentity()]

    @factory.post_generation
    def CUSOFFSENT740(self, create, extracted, **kwargs):
        if extracted:
            # A list of groups were passed in, use them
            self['CUSOFFSENT740'] = extracted
        else:
            self['CUSOFFSENT740'] = [CustomsOfficeSubsequentEntry()]


class ENSPresentationFactory(BaseMessageMixin, factory.Factory):
    class Meta:
        model = dict


class ENSModificationFactory(BaseMessageMixin, factory.Factory):
    class Meta:
        model = dict

    NOTPAR670 = factory.SubFactory(NotifyParty)
    HEAHEA = factory.SubFactory(ENSModificationHeader)



import logging

from . import zeep_plugins
from .wsdl import ADUANET_SERVICES
from requests import Session
from zeep import exceptions as zeep_exceptions
from zeep import Client, Transport, xsd


logger = logging.getLogger(__name__)


class Config:
    def __init__(self, service_name, test_mode):
        self.payload = ADUANET_SERVICES[service_name]
        self.test_mode = test_mode

    def __getattr__(self, name):
        if name == 'port':
            test, prod = self.payload['port_test'], self.payload['port_production']
            return test if self.test_mode else prod

        return self.payload[name]

    def __str__(self):
        return f'''
        Config: {self.verbose_name}
        WSDL: {self.wsdl}
        Operation: {self.operation}
        Port: {self.port}
        Service: {self.service}
        '''


class Result:
    default_context = {
        'raw_request': None,
        'raw_response': None,
        'exception': None,
    }

    def __init__(self, data, error, context=None):
        self.context = context or {}

        for k, v in self.default_context.items():
            self.context[k] = self.context.get(k, v)

        self.data = data
        self.error = error

    def __getattr__(self, name):
        return self.context[name]

    @property
    def valid(self):
        return self.error is None


class Controller:
    def __init__(self, client, config, raw_xml_plugin=None):
        self.client = client
        self.config = config
        self.raw_xml_plugin = raw_xml_plugin
        logger.info('Controller initialized with config %s', config)

    @classmethod
    def build_from_config(cls, config, pub_cert, priv_cert):
        '''
        Builds AEAT Controller with Plugins

        :param config: Preconfigured Config object
        :param pub_cert: Public certificate file path
        :param priv_cert: Private certificate file path

        :rtype: Controller
        '''
        session = Session()
        session.cert = (pub_cert, priv_cert)
        transport = Transport(session=session, operation_timeout=60)

        raw_xml_plugin = zeep_plugins.RawXMLPlugin()

        if config.signed:
            sign_plugin = zeep_plugins.SignMessagePlugin(pub_cert, priv_cert)
            plugins = [raw_xml_plugin, sign_plugin]
        else:
            plugins = [raw_xml_plugin]

        client = Client(config.wsdl, service_name=config.service, port_name=config.port,
                        transport=transport, strict=False, plugins=plugins)

        return cls(client, config, raw_xml_plugin=raw_xml_plugin)

    @property
    def operation(self):
        return getattr(self.client.service, self.config.operation)

    def request(self, payload):
        if self.config.signed:
            # Skip WSDL validation. Is added later by zeep_plugins.SignMessagePlugin
            payload['Signature'] = xsd.SkipValue

        try:
            data = self.operation(**payload)
        except zeep_exceptions.Fault as e:
            logger.info('AEAT request failed.', exc_info=True)
            result = Result(None, e.message, context={'exception': e.message})
        except zeep_exceptions.Error as e:
            logger.info('AEAT request failed.', exc_info=True)
            result = Result(None, 'Wrong AEAT response',  context={'exception': e.message})
        except Exception as e:
            logger.critical('Unexpected exception', exc_info=True)
            result = Result(None, 'Unknown error',  context={'exception': str(e)})
        else:
            result = Result(data, None)

        if self.raw_xml_plugin:
            result.context['raw_request'] = self.raw_xml_plugin.last_sent
            result.context['raw_response'] = self.raw_xml_plugin.last_received

        return result



import pytest

import setup_django  # NOQA

from aeat.rest_framework import serializers


def test_ens_serializer(zeep_response):
    aeat_response = zeep_response(
        'enswsv5',
        'ens_presentation_IE315V5.wsdl', 'ens_presentation_success_IE328V5Sal.xml', 'IE315V5'
    )

    serializer = serializers.ENSSerializer(data=aeat_response)
    assert serializer.is_valid(raise_exception=False), serializer.errors

    assert {'mrn': '17ES004311Z0000010'} == serializer.data


def test_exs_serializer(zeep_response):
    aeat_response = zeep_response(
        'exswsv2', 'exs_presentation_IE615V2.wsdl', 'exs_presentation_success_IE628V2Sal.xml',
        'IE615V2'
    )

    serializer = serializers.EXSSerializer(data=aeat_response)
    assert serializer.is_valid(raise_exception=False), serializer.errors

    assert {'mrn': '17ES00361160001234',
            'item_number_involved': 0, 'customs_intervention_code': 'V'} == serializer.data


@pytest.mark.parametrize('ver,url,response,operation,expected,is_error,expected_data', [
    # ENS
    ('enswsv5', 'ens_presentation_IE315V5.wsdl', 'ens_presentation_success_IE328V5Sal.xml',
     'IE315V5', serializers.ENSSerializer, False, {'mrn': '17ES004311Z0000010'}),

    ('enswsv5', 'ens_modification_IE313V5.wsdl', 'ens_modification_success_IE304V5Sal.xml',
     'IE313V5', serializers.ENSSerializer, False, {'mrn': '18ES003611Z0123456'}),

    ('enswsv5', 'ens_presentation_IE315V5.wsdl', 'ens_presentation_error_IE316V5Sal.xml',
     'IE315V5', serializers.ENSFunctionalErrorSerializer, True,
     {'type': '12', 'pointer': 'MES.MesSenMES3', 'code': '',
      'reason': '1234-Message Sender is not valid'}),

    ('enswsv5', 'ens_presentation_IE315V5.wsdl', 'ens_presentation_error_IE316V5Sal_v2.xml',
     'IE315V5', serializers.ENSFunctionalErrorSerializer, True,
     {'type': '12', 'pointer': 'ITI.ITI', 'code': 'R879',
      'reason': 'No error message supplied'}),

    # EXS
    ('exswsv2', 'exs_presentation_IE615V2.wsdl', 'exs_presentation_success_IE628V2Sal.xml',
     'IE615V2', serializers.EXSSerializer, False,
     {'mrn': '17ES00361160001234', 'customs_intervention_code': 'V', 'item_number_involved': 0}),

    ('exswsv2', 'exs_presentation_IE615V2.wsdl', 'exs_presentation_error_IE919V2Sal.xml',
     'IE615V2', serializers.UnknownResponseSerializer, True, {'reason': 'Unknown AEAT response'}),
])
def test_get_serializer_for_mapped_response(ver, zeep_response, url, response, operation, expected,
                                            is_error, expected_data):
    aeat_response = zeep_response(ver, url, response, operation)
    serializer_cls = serializers.get_class_for_aeat_response(aeat_response)
    assert expected == serializer_cls

    # It should parse the response without errors
    serializer = serializer_cls(data=aeat_response)
    assert is_error == serializer.is_error
    assert serializer.is_valid(), serializer.errors
    assert expected_data == serializer.data


def test_get_serializer_for_unmapped_response():
    aeat_response = {}
    serializer_cls = serializers.get_class_for_aeat_response(aeat_response)
    assert serializers.UnknownResponseSerializer == serializer_cls
    serializer = serializer_cls(data=aeat_response)
    assert serializer.is_valid(), serializer.errors
    assert {'reason': 'Unknown AEAT response'} == serializer.data


@pytest.mark.parametrize('filename,expected', [
    ('exs_presentation_success_IE628V2Sal.xml',
     'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/adrx/ws/IE628V2Sal.xsd'),

    ('exs_presentation_error_IE919V2Sal.xml',
     'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/adrx/ws/IE919V2Sal.xsd'),

    ('ens_presentation_success_IE328V5Sal.xml',
     'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/aden/enswsv5/IE328V5Sal.xsd'),  # NOQA

    ('ens_modification_success_IE304V5Sal.xml',
     'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/aden/enswsv5/IE304V5Sal.xsd'),  # NOQA

    ('ens_presentation_error_IE917V5Sal.xml',
     'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/aden/enswsv5/IE917V5Sal.xsd'),  # NOQA
])
def test_parse_xsd(response_etree_element, filename, expected):
    xml = response_etree_element(filename)
    assert expected == serializers.parse_xsd(xml)



import pytest

import factories_v2
import factories_v5
import setup_django  # NOQA

import aeat
import aeat.rest_framework.validators as validators


'''
Test tagged as functional are are not run by default (see setup.cfg)
Copy .env-sample to .env and set the certificate paths

Enable the env vars with ```export `cat .env````
then run pytest specifiying the tag ```pytest -m functional```

Be aware that they are run against the AEAT (in test mode) with your REAL certificate.
'''


@pytest.fixture
def make_aeat_test_controller(certificate_real):
    def make_controller_for_service(service_name):
        config = aeat.Config(service_name, test_mode=True)
        return aeat.Controller.build_from_config(config, *certificate_real)

    return make_controller_for_service


@pytest.mark.functional
def test_ens_presentation(make_aeat_test_controller):
    payload = factories_v5.ENSPresentationFactory(MesIdeMES19='TEST10002')
    serializer = validators.ENSPresentationValidator(data=payload)
    assert serializer.is_valid(raise_exception=False), serializer.errors

    controller = make_aeat_test_controller('ens_presentation')
    result = controller.request(serializer.data)

    assert result.valid, f'Error: {result.error} | Raw \n: {result.raw_response}'
    assert result.data


@pytest.mark.functional
def test_ens_modification(make_aeat_test_controller):
    payload = factories_v5.ENSModificationFactory(MesIdeMES19='TEST20001')
    serializer = validators.ENSModificationValidator(data=payload)
    assert serializer.is_valid(raise_exception=False), serializer.errors

    controller = make_aeat_test_controller('ens_modification')
    result = controller.request(serializer.data)

    assert result.valid, f'Error: {result.error} | Raw \n: {result.raw_response}'
    assert result.data


@pytest.mark.functional
def test_exs_presentation(make_aeat_test_controller):
    payload = factories_v2.EXSPresentationFactory(MesIdeMES19='TEST30005')
    serializer = validators.EXSPresentationValidator(data=payload)
    assert serializer.is_valid(raise_exception=False), serializer.errors

    controller = make_aeat_test_controller('exs_common')
    result = controller.request(serializer.data)

    assert result.valid, f'Error: {result.error} | Raw \n: {result.raw_response}'
    assert result.data


@pytest.mark.functional
def test_exs_modification(make_aeat_test_controller):
    payload = factories_v2.EXSModificationFactory(MesIdeMES19='TEST30003')
    serializer = validators.EXSModificationValidator(data=payload)
    assert serializer.is_valid(raise_exception=False), serializer.errors

    controller = make_aeat_test_controller('exs_common')
    result = controller.request(serializer.data)

    assert result.valid, f'Error: {result.error} | Raw \n: {result.raw_response}'
    assert result.data


@pytest.mark.functional
def test_exs_cancellation(make_aeat_test_controller):
    payload = factories_v2.EXSCancellationFactory(MesIdeMES19='TEST30003')
    serializer = validators.EXSCancellationValidator(data=payload)
    assert serializer.is_valid(raise_exception=False), serializer.errors

    controller = make_aeat_test_controller('exs_common')
    result = controller.request(serializer.data)

    assert result.valid, f'Error: {result.error} | Raw \n: {result.raw_response}'
    assert result.data



from django.conf import settings
from rest_framework import serializers as rf

from . import complex_types_v2 as v2
from . import complex_types_v5 as v5
from . import fields


class ValidatorBase(rf.Serializer):
    service_name = None
    include_test_mode = True

    def to_representation(self, instance):
        instance = super().to_representation(instance)

        if self.include_test_mode and settings.AEAT_TEST_MODE:
            instance['TesIndMES18'] = '0'

        return instance


class ENSForkValidator(ValidatorBase):
    service_name = 'ens_fork'


class ENSPresentationValidator(v5.BaseMixin, ValidatorBase):
    service_name = 'ens_presentation'

    MesTypMES20 = fields.NotRequiredStr(default='CC315A', read_only=True,
                                        help_text='Message type. EG CC315A')
    HEAHEA = v5.ENSPresentationHeader(required=True)


class ENSModificationValidator(v5.BaseMixin, ValidatorBase):
    service_name = 'ens_modification'
    NOTPAR670 = v5.NotifyParty(required=False)
    MesTypMES20 = fields.NotRequiredStr(default='CC313A', read_only=True,
                                        help_text='Message type. EG CC313A')
    HEAHEA = v5.ENSModificationHeader(required=True)


class EXSPresentationValidator(v2.BaseMixin, ValidatorBase):
    service_name = 'exs_common'

    HEAHEA = v2.EXSHeader(required=True)


class EXSModificationValidator(v2.BaseMixin, ValidatorBase):
    service_name = 'exs_common'

    HEAHEA = v2.EXSHeaderModification(required=True)


class EXSCancellationValidator(v2.BaseMixin, ValidatorBase):
    service_name = 'exs_common'

    HEAHEA = v2.EXSHeaderCancellation(required=True)



from rest_framework import serializers as rf


class AEATDateField(rf.DateField):
    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **kwargs, format='%y%m%d')


class AEATDateTimeField(rf.DateTimeField):
    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **kwargs, format='%Y%m%d%H%M')


class AEATTimeField(rf.TimeField):
    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **kwargs, format='%H%M')


class RequiredStr(rf.CharField):
    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **kwargs, required=True)


class NotRequiredStr(rf.CharField):
    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **kwargs, required=False)



import pytest

from lxml import etree
from zeep.wsdl import utils

from aeat import zeep_plugins


@pytest.fixture
def envelope():
    envelope = etree.Element('Envelope')
    body = etree.SubElement(envelope, 'Body')
    etree.SubElement(body, 'Message', Id='myid')
    return envelope


def test_sign_message_egress(certificate_example, envelope):
    plugin = zeep_plugins.SignMessagePlugin(*certificate_example)
    envelope, _ = plugin.egress(envelope, None, 'test', None)

    xml_str = utils.etree_to_string(envelope).decode()

    assert '<Signature xmlns="http://www.w3.org/2000/09/xmldsig#" Id="Firma">' in xml_str


def test_raw_xml_plugin(envelope):
    plugin = zeep_plugins.RawXMLPlugin()

    plugin.ingress(envelope, None, None)
    plugin.egress(envelope, None, None, None)

    expected = '<Envelope>\n  <Body>\n    <Message Id="myid"/>\n  </Body>\n</Envelope>\n'
    assert expected == plugin.last_sent == plugin.last_received



BASE_ENS = 'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/aden/'

ADUANET_SERVICES = {
    'ens_fork': {
        'verbose_name': 'Servicio Solicitud Desvío ENS V5.0',
        'wsdl': f'{BASE_ENS}enswsv5/IE323V5.wsdl',
        'operation': 'IE323V5',
        'port_production': 'IE323V5',
        'port_test': 'IE323V5Pruebas',
        'service': 'IE323V5Service',
        'signed': True,
    },
    'ens_modification': {
        'verbose_name': 'Servicio Modificación ENS V5.0',
        'wsdl': f'{BASE_ENS}enswsv5/IE313V5.wsdl',
        'operation': 'IE313V5',
        'port_production': 'IE313V5',
        'port_test': 'IE313V5Pruebas',
        'service': 'IE313V5Service',
        'signed': True,
    },
    'ens_presentation': {
        'verbose_name': 'Servicio de Presentación ENS V5.0',
        'wsdl': f'{BASE_ENS}enswsv5/IE315V5.wsdl',
        'operation': 'IE315V5',
        'port_production': 'IE315V5',
        'port_test': 'IE315V5Pruebas',
        'service': 'IE315V5Service',
        'signed': True,
    },

    'exs_common': {
        'verbose_name': "Servicio de envío de EXIT Sumary (EXS) v2",
        'wsdl': 'https://www2.agenciatributaria.gob.es/ADUA/internet/es/aeat/dit/adu/adrx/ws/IE615V2.wsdl',  # NOQA
        'operation': 'IE615V2',
        'port_production': 'IE615V2',
        'port_test': 'IE615V2Pruebas',
        'service': 'IE615V2Service',
        'signed': True,
    },
}




__title__ = 'aeat-web-services'
__summary__ = 'Spanish Tax Agency Electronic Office (AEAT) Integration'
__version__ = '2.0.2'
__license__ = 'MIT'
__uri__ = 'https://github.com/initios/aeat-web-services'
__author__ = 'Initios Desarrollos'
__email__ = 'dev@initios.com'



import logging

from lxml import etree
from zeep import Plugin

from aeat import utils, xml_signing

logger = logging.getLogger(__name__)


class SignMessagePlugin(Plugin):
    def __init__(self, cert_path, key_path):
        self.cert_path = cert_path
        self.key_path = key_path

    def egress(self, envelope, http_headers, operation, binding_options):
        '''Sign enveloped message'''
        args = envelope[0][0], self.cert_path, self.key_path
        signed = xml_signing.sign(*args)

        data = '''
        <soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/">
        <soapenv:Header/>
        <soapenv:Body>{}</soapenv:Body>
        </soapenv:Envelope>'''.format(
            etree.tostring(signed).decode())

        envelope = etree.fromstring(data)
        return envelope, http_headers


class RawXMLPlugin(object):
    '''
    Stores last request and response as str
    '''
    def __init__(self):
        self.last_sent = None
        self.last_received = None

    def ingress(self, envelope, http_headers, operation):
        self.last_received = utils.lxml_to_string(envelope)
        logging.info(self.last_received)
        return envelope, http_headers

    def egress(self, envelope, http_headers, operation, binding_options):
        self.last_sent = utils.lxml_to_string(envelope)
        logging.info(self.last_sent)
        return envelope, http_headers



from django.test.utils import override_settings
from rest_framework import serializers as rf

from aeat.rest_framework import validators


def test_adds_test_indicator_when_test_mode_is_enabled():
    class TestSerializer(validators.ValidatorBase, rf.Serializer):
        pass

    serializer = TestSerializer(data={})
    serializer.is_valid(raise_exception=True)

    assert '0' == serializer.data['TesIndMES18']


@override_settings(AEAT_TEST_MODE=False)
def test_does_not_adds_test_indicator_in_production():
    class TestSerializer(validators.ValidatorBase, rf.Serializer):
        pass

    serializer = TestSerializer(data={})
    serializer.is_valid(raise_exception=True)

    assert 'TesIndMES18' not in serializer.data



#!/usr/bin/env python
# Copyright 2018 Initios Desarrollos
#
# All rights reserved

import os

import setuptools

base_dir = os.path.dirname(__file__)

about = {}

with open(os.path.join(base_dir, 'src', 'aeat', '__about__.py')) as f:
    exec(f.read(), about)

with open(os.path.join(base_dir, 'README.rst')) as f:
    long_description = f.read()


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# https://packaging.python.org/tutorials/distributing-packages/

setuptools.setup(
    name=about['__title__'],
    version=about['__version__'],

    description=about['__summary__'],
    long_description=long_description,
    license=about['__license__'],
    url=about['__uri__'],

    author=about['__author__'],
    author_email=about['__email__'],

    python_requires='>=3.4,<3.7',

    install_requires=[
        'lxml>=4.1.1,<4.2.0',
        'xmlsec>=1.3.3,<1.4.0',
        'zeep>=2.4.0,<2.5.0',
    ],

    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
    ],

    packages=['aeat', 'aeat.rest_framework'],
    package_dir={'': 'src'},
)



from lxml import etree

from aeat import xml_signing


def test_sign_without_root_id(certificate_example):
    root = etree.Element('Data')

    signed = xml_signing.sign(root, *certificate_example)

    assert xml_signing.verify(signed, *certificate_example)
    assert 'MessageRoot' == signed.attrib['Id']


def test_verify_simple_signature(certificate_example):
    root = etree.Element('Data', Id='myid')
    signed = xml_signing.sign(root, *certificate_example)

    assert xml_signing.verify(signed, *certificate_example)


def test_verify_wrong_signature(certificate_example):
    root = etree.Element('Data', Id='myid')

    signed = xml_signing.sign(root, *certificate_example)
    etree.SubElement(signed, 'KeyAfterSigning')

    assert xml_signing.verify(signed, *certificate_example) is False


def test_sign_complete_xades_epel(certificate_example):
    root = etree.Element('Data', Id='myid')

    signed = xml_signing.sign(root, *certificate_example)
    assert xml_signing.verify(signed, *certificate_example)



import datetime as dt
import logging
from datetime import timezone as tz

import xmlsec
from lxml import etree

logger = logging.getLogger(__name__)


def aeat_object_with_qualifying_properties():
    now_iso_format = dt.datetime.now(tz.utc).astimezone().isoformat(timespec='seconds')

    return etree.fromstring(f'''
        <Object xmlns="http://www.w3.org/2000/09/xmldsig#">
            <etsi:QualifyingProperties
                xmlns:ds="http://www.w3.org/2000/09/xmldsig#"
                xmlns:etsi="http://uri.etsi.org/01903/v1.2.2#" Target="#Firma">
                <etsi:SignedProperties Id="SignatureProperties">
                    <etsi:SignedSignatureProperties>
                        <etsi:SigningTime>{now_iso_format}</etsi:SigningTime>
                        <etsi:SignaturePolicyIdentifier>
                            <etsi:SignaturePolicyId>
                                <etsi:SigPolicyId>
                                    <etsi:Identifier>http://administracionelectronica.gob.es/es/ctt/politicafirma/politica_firma_AGE_v1_8.pdf</etsi:Identifier>
                                </etsi:SigPolicyId>
                                <etsi:SigPolicyHash>
                                    <ds:DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"/>
                                    <ds:DigestValue>VYICYpNOjso9g1mBiXDVxNORpKk=</ds:DigestValue>
                                </etsi:SigPolicyHash>
                            </etsi:SignaturePolicyId>
                        </etsi:SignaturePolicyIdentifier>
                    </etsi:SignedSignatureProperties>
                </etsi:SignedProperties>
            </etsi:QualifyingProperties>
        </Object>
        ''')  # NOQA


def sign(root, cert, key):
    '''
    Sign LXMLElement node with specified certificate
    '''
    root_id = root.get('Id')

    if not root_id:
        root.attrib['Id'] = 'MessageRoot'
        root_id = 'MessageRoot'

    elem_uri = "#%s" % root_id
    signature = xmlsec.template.create(root,
                                       c14n_method=xmlsec.constants.TransformInclC14NWithComments,
                                       sign_method=xmlsec.constants.TransformRsaSha1)
    signature.attrib['Id'] = 'Firma'
    # Main reference
    ref = xmlsec.template.add_reference(signature,
                                        xmlsec.constants.TransformSha1,
                                        uri=elem_uri)
    xmlsec.template.add_transform(ref, xmlsec.constants.TransformEnveloped)

    # Keyinfo reference
    ref = xmlsec.template.add_reference(signature,
                                        xmlsec.constants.TransformSha1,
                                        uri='#CertificadoFirmante')
    # SignedProperties reference
    ref = xmlsec.template.add_reference(signature,
                                        xmlsec.constants.TransformSha1,
                                        uri='#SignatureProperties')

    # Add x509 keyinfo
    ki = xmlsec.template.ensure_key_info(signature, id="CertificadoFirmante")
    xmlsec.template.add_x509_data(ki)

    # Add external object
    signature.append(aeat_object_with_qualifying_properties())
    root.append(signature)

    ctx = xmlsec.SignatureContext()
    ctx.register_id(root, id_attr="Id")

    ctx.key = xmlsec.Key.from_file(key,
                                   format=xmlsec.constants.KeyDataFormatPem)
    ctx.key.load_cert_from_file(cert, xmlsec.constants.KeyDataFormatPem)

    ctx.sign(signature)
    return root


def verify(root, cert, key):
    signed = xmlsec.tree.find_node(root, xmlsec.constants.NodeSignature)
    ctx1 = xmlsec.SignatureContext()
    ctx1.key = xmlsec.Key.from_file(key,
                                    format=xmlsec.constants.KeyDataFormatPem)
    ctx1.key.load_cert_from_file(cert,
                                 xmlsec.constants.KeyDataFormatPem)

    try:
        ctx1.verify(signed)
    except xmlsec.Error:
        logger.error("Invalid signature", exc_info=True)
        return False
    else:
        return True
